/*
 * deviced
 *
 * Copyright (c) 2012 - 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include <errno.h>
#include <dlfcn.h>

#include "ambient-mode.h"
#include "core/log.h"
#include "core/devices.h"
#include "core/common.h"
#include "core/device-notifier.h"
#include "util.h"
#include "device-interface.h"
#include "vconf.h"
#include "core.h"
#include "device-node.h"
#include "display/display-dpms.h"
#include "battery-monitor.h"
#include "battery/power-supply.h"

#define SET_SUSPEND_TIME		0.5

#define TOUCH_ON	1
#define TOUCH_OFF	0

#define LCD_PHASED_MIN_BRIGHTNESS	1
#define LCD_PHASED_MAX_BRIGHTNESS	100
#define LCD_PHASED_CHANGE_STEP		5
#define LCD_PHASED_DELAY		10000 	/* microsecond */
#define DUMP_MODE_WAITING_TIME		600000	/* milisecond */

#define POWER_AUTOSLEEP_PATH    "/sys/power/autosleep"
#define POWER_LOCK_PATH         "/sys/power/wake_lock"
#define POWER_UNLOCK_PATH       "/sys/power/wake_unlock"
#define POWER_WAKEUP_PATH       "/sys/power/wakeup_count"
#define POWER_STATE_PATH        "/sys/power/state"

#define DISPLAY_HAL_LIB_PATH "/usr/lib/libdisplay-hal.so"

#define GESTURE_STR			"gesture"
#define POWERKEY_STR			"powerkey"
#define EVENT_STR			"event"
#define TOUCH_STR			"touch"
#define BEZEL_STR			"bezel"
#define TIMEOUT_STR			"timeout"
#define PROXIMITY_STR			"proximity"
#define PALM_STR			"palm"
#define UNKNOWN_STR			"unknown"

#define FREEZER_VITAL_WAKEUP_CGROUP "/sys/fs/cgroup/freezer/vital_wakeup/freezer.state"

struct _backlight_ops backlight_ops;
struct _power_ops power_ops;

static int mainlock_status = POWER_UNLOCK;
static bool custom_status;
static int custom_brightness;
static int force_brightness;
static int default_brightness;
static int vital_support = -2;
static int vital_service;
static bool vital_sleep;
static int dpms_running_state = DPMS_SETTING_DONE;
static struct display_device *display_dev;
static guint release_timer;
static int aod_max_level = -1;
static int aod_normal_level = -1;
static int aod_min_level = -1;

struct display_device *display_dev_get(void)
{
	return display_dev;
}

void dpms_set_running_state(int val)
{
	dpms_running_state = val;
}

static int bl_onoff(int on, enum device_flags flags)
{
	dpms_set_state(on);

#ifdef ENABLE_PM_LOG
	if (on == DPMS_ON)
		pm_history_save(PM_LOG_LCD_ON_COMPLETE, pm_cur_state);
	else if (on == DPMS_OFF || on == DPMS_FORCE_OFF)
		pm_history_save(PM_LOG_LCD_OFF_COMPLETE, pm_cur_state);
	else
		pm_history_save(PM_LOG_LCD_CONTROL_FAIL, on);
#endif

	return 0;
}

static int bl_brt(int brightness, int delay)
{
	int ret = -1;

	if (delay > 0)
		usleep(delay);

	/* Update device brightness */
	ret = backlight_ops.set_brightness(brightness);

	_I("Set brightness(%d): %d", brightness, ret);

	return ret;
}

static int system_suspend(void)
{
	int ret;

	_I("system suspend");
	ret = sys_set_str(POWER_STATE_PATH, "mem");
	_I("System resume: %d", ret);
	return 0;
}

static int system_enable_autosleep(void)
{
	_I("System autosleep enabled.");
	return sys_set_str(POWER_AUTOSLEEP_PATH, "mem");
}

static int vital_mode_support(void)
{
	if (vital_support < 0) {
		FILE *fp;

		fp = fopen(FREEZER_VITAL_WAKEUP_CGROUP, "r");
		if (fp == NULL) {
			_E("%s open failed", FREEZER_VITAL_WAKEUP_CGROUP);
			/* read max 2 times to check if this file exist */
			vital_support++;
			return 0;
		}
		vital_support = 1;
		fclose(fp);
	}
	return vital_support;
}

static int suspend_other_process(int type)
{
	int ret = 0;
	char buf[8];
	const char *command[1];

	if (vital_service == type)
		return ret;

	if (type == VITAL_WAKEUP && vital_service > VITAL_SLEEP)
		return ret;

	vital_service = type;

	if (!vital_mode_support())
		return ret;

	if (type == VITAL_SLEEP) {
		snprintf(buf, sizeof(buf), "%s", "sleep");
		command[0] = buf;
		dbus_handle_method_sync_timeout(RESOURCED_BUS_NAME,
						RESOURCED_PATH_FREEZER,
						RESOURCED_INTERFACE_FREEZER,
						"SetSuspend",
						"s",
						command,
						SET_SUSPEND_TIME*1000);
		vital_sleep = true;
	} else if (type == VITAL_WAKEUP) {
		snprintf(buf, sizeof(buf), "%s", "wakeup");
		command[0] = buf;
		ret = dbus_handle_method_async(RESOURCED_BUS_NAME,
						RESOURCED_PATH_FREEZER,
						RESOURCED_INTERFACE_FREEZER,
						"SetSuspend",
						"s", command);
	} else if (type == VITAL_EXIT) {
		snprintf(buf, sizeof(buf), "%s", "exit");
		command[0] = buf;
		ret = dbus_handle_method_async(RESOURCED_BUS_NAME,
						RESOURCED_PATH_FREEZER,
						RESOURCED_INTERFACE_FREEZER,
						"SetSuspend",
						"s", command);
		vital_sleep = false;
	}
	return ret;
}

static int system_power_lock(void)
{
	_I("system power lock");
	suspend_other_process(VITAL_WAKEUP);
	mainlock_status = POWER_LOCK;

	return sys_set_str(POWER_LOCK_PATH, "mainlock");
}

static int system_power_unlock(void)
{
	_I("system power unlock");
	suspend_other_process(VITAL_SLEEP);
	mainlock_status = POWER_UNLOCK;

	return sys_set_str(POWER_UNLOCK_PATH, "mainlock");
}

static int system_get_power_lock(void)
{
	return mainlock_status;
}

static int system_get_power_lock_support(void)
{
	static int power_lock_support = -1;
	int ret;

	if (power_lock_support >= 0)
		goto out;

	ret = sys_check_node(POWER_LOCK_PATH);
	if (ret < 0)
		power_lock_support = false;
	else
		power_lock_support = true;

	_I("System power lock: %s",
			(power_lock_support ? "support" : "not support"));

out:
	return power_lock_support;
}

static int get_lcd_power(void)
{
	enum display_state val;
	int ret;

	if (!display_dev || !display_dev->get_state) {
		return dpms_get_state();
	} else {
		ret = display_dev->get_state(&val);
		if (ret < 0)
			return ret;

		if (val == DISPLAY_ON && ambient_get_state())
			return DPMS_OFF;

		if (dpms_running_state != DPMS_SETTING_DONE)
			return dpms_running_state;

		switch (val) {
		case DISPLAY_ON:
			return DPMS_ON;
		case DISPLAY_STANDBY:
			return DPMS_STANDBY;
		case DISPLAY_SUSPEND:
			return DPMS_SUSPEND;
		case DISPLAY_OFF:
			return DPMS_OFF;
		case DISPLAY_DETACH:
			return DPMS_DETACH;
		default:
			return -EINVAL;
		}
	}
}

bool display_dimstay_check(void)
{
	if (pm_status_flag & DIM_FLAG)
		return true;

	if ((pm_status_flag & PWRSV_FLAG) && !(pm_status_flag & BRTCH_FLAG))
		return true;

	return false;
}

static void change_brightness(int start, int end, int step)
{
	int diff, val;
	int ret = -1;
	int prev;

	if (display_dimstay_check())
		return;

	ret = backlight_ops.get_brightness(&prev);
	if (ret < 0) {
		_E("Failed to get brightness: %d", ret);
		return;
	}

	if (prev == end)
		return;

	if (pm_status_flag & DIM_MASK)
		end = 0;

	_I("start %d end %d step %d", start, end, step);

	if (display_dev && display_dev->set_multi_brightness) {
		ret = display_dev->set_multi_brightness(end, step, LCD_PHASED_DELAY);
		if (ret < 0)
			_E("Failed to set_multi_brightness (%d)", ret);

		backlight_ops.set_brightness(end);

		return;
	}

	diff = end - start;

	if (abs(diff) < step)
		val = (diff > 0 ? 1 : -1);
	else
		val = (int)ceil((double)diff / step);

	while (start != end) {
		if (val == 0) break;

		start += val;
		if ((val > 0 && start > end) ||
		    (val < 0 && start < end))
			start = end;

		bl_brt(start, LCD_PHASED_DELAY);
	}
}

static int backlight_on(enum device_flags flags)
{
	int ret = -1;
	static int cnt;

	_I("[DPMS XLIB Backlight] LCD on %#x cnt:%d", flags, cnt);

	cnt++;
	ret = bl_onoff(DPMS_ON, flags);
#ifdef ENABLE_PM_LOG
	pm_history_save(PM_LOG_LCD_ON, pm_cur_state);
#endif

	return ret;
}

static int backlight_off(enum device_flags flags)
{
	int ret = -1;
	static int cnt, ambient_cnt;

	if (flags & AMBIENT_MODE) {
		_I("[DPMS XLIB Backlight] LCD suspend %#x cnt:%d", flags, ambient_cnt);
		ambient_cnt++;

		return 0;
	}

	_I("[DPMS XLIB Backlight] LCD off %#x cnt:%d", flags, cnt);
	cnt++;

	if (flags & LCD_PHASED_TRANSIT_MODE)
		backlight_ops.transit_brt(default_brightness,
		    LCD_PHASED_MIN_BRIGHTNESS, LCD_PHASED_CHANGE_STEP);

	if (flags & FORCE_OFF_MODE)
		ret = bl_onoff(DPMS_FORCE_OFF, flags);
	else
		ret = bl_onoff(DPMS_OFF, flags);

#ifdef ENABLE_PM_LOG
	pm_history_save(PM_LOG_LCD_OFF, pm_cur_state);
#endif

	return ret;
}

static int backlight_dim(void)
{
	int ret;

	ret = backlight_ops.set_brightness(PM_DIM_BRIGHTNESS);
#ifdef ENABLE_PM_LOG
	if (!ret)
		pm_history_save(PM_LOG_LCD_DIM, pm_cur_state);
	else
		pm_history_save(PM_LOG_LCD_DIM_FAIL, pm_cur_state);
#endif
	return ret;
}

static int set_custom_status(bool on)
{
	custom_status = on;
	return 0;
}

static bool get_custom_status(void)
{
	return custom_status;
}

static int save_custom_brightness(void)
{
	int ret, brightness;

	ret = backlight_ops.get_brightness(&brightness);

	custom_brightness = brightness;

	return ret;
}

static int custom_backlight_update(void)
{
	int ret = 0;

	if (custom_brightness < PM_MIN_BRIGHTNESS ||
	    custom_brightness > PM_MAX_BRIGHTNESS)
		return -EINVAL;

	if (display_dimstay_check())
		ret = backlight_dim();
	else {
		_I("custom brightness restored! %d", custom_brightness);
		ret = backlight_ops.set_brightness(custom_brightness);
	}

	return ret;
}

static int set_force_brightness(int level)
{
	if (level < 0 ||  level > PM_MAX_BRIGHTNESS)
		return -EINVAL;

	force_brightness = level;

	return 0;
}

static int backlight_update(void)
{
	int ret = 0;
	int brt;

	if (get_custom_status()) {
		_I("custom brightness mode! brt no updated");
		return 0;
	}
	if (display_dimstay_check())
		ret = backlight_dim();
	else {
		brt = backlight_ops.get_default_brt();
		ret = backlight_ops.set_brightness(brt);
	}
	return ret;
}

static int backlight_standby(int force)
{
	int ret = -1;

	if ((get_lcd_power() == DPMS_ON) || force) {
		_I("LCD standby");
		ret = bl_onoff(DPMS_STANDBY, 0);
	}

	return ret;
}

static int set_default_brt(int level)
{
	if (level < PM_MIN_BRIGHTNESS || level > PM_MAX_BRIGHTNESS)
		level = display_conf.pm_default_brightness;

	default_brightness = level;

	return 0;
}

static int get_default_brt(void)
{
	return default_brightness;
}

static int check_wakeup_src(void)
{
	/*  TODO if nedded.
	 * return wackeup source. user input or device interrupts? (EVENT_DEVICE or EVENT_INPUT)
	 */
	return EVENT_DEVICE;
}

static int get_wakeup_count(int *cnt)
{
	int ret;
	int wakeup_count;

	if (!cnt)
		return -EINVAL;

	ret = sys_get_int(POWER_WAKEUP_PATH, &wakeup_count);
	if (ret < 0)
		return ret;

	*cnt = wakeup_count;
	return 0;
}

static int set_wakeup_count(int cnt)
{
	int ret;

	ret = sys_set_int(POWER_WAKEUP_PATH, cnt);
	if (ret < 0)
		return ret;

	return 0;
}

static int get_max_brightness(void)
{
	static int max = -1;
	int ret;

	if (max > 0)
		return max;

	if (!display_dev) {
		_E("There is no display device.");
		return -ENOENT;
	}

	if (!display_dev->get_max_brightness) {
		max = DEFAULT_DISPLAY_MAX_BRIGHTNESS;
		return max;
	}

	ret = display_dev->get_max_brightness(&max);
	if (ret < 0) {
		_E("Failed to get max brightness: %d", ret);
		return ret;
	}

	return max;
}

static int set_brightness(int val)
{
	int max;

	if (!display_dev || !display_dev->set_brightness) {
		_E("There is no display device.");
		return -ENOENT;
	}

	max = get_max_brightness();
	if (max < 0) {
		_E("Failed to get max brightness.");
		return max;
	}

	if (force_brightness > 0 && val != PM_DIM_BRIGHTNESS) {
		_I("brightness(%d), force brightness(%d)",
		    val, force_brightness);
		val = force_brightness;
	}

	if (pm_status_flag & DIM_MASK)
		val = 0;

	/* Maximum Brightness to users is 100.
	 * Thus real brightness need to be calculated */
	val = val * max / 100;

	update_bds_brightness_record(val);
	_I("set brightness %d (default:%d)", val, default_brightness);
	device_notify(DEVICE_NOTIFIER_DISPLAY_BRIGHTNESS, (void *)&val);

	return display_dev->set_brightness(val);
}

static int get_brt_normalized(int brt_raw)
{
	int quotient, remainder;
	int max;

	max = get_max_brightness();
	if (max < 0) {
		_E("Failed to get max brightness.");
		return max;
	}

	/* Maximum Brightness to users is 100.
	 * Thus the brightness value need to be calculated using real brightness.
	 *    ex) Let's suppose that the maximum brightness of driver is 255.
	 *      case 1) When the user sets the brightness to 41,
	 *              real brightness is
	 *                 41 * 255 / 100 = 104.55 = 104 (rounded off)
	 *      case 2) When the user gets the brightness,
	 *              the driver returns the brightness 104.
	 *              Thus the brightness to users is
	 *                 104 * 100 / 255 = 40.7843.... = 41 (rounded up)
	 */
	quotient = brt_raw * 100 / max;
	remainder = brt_raw * 100 % max;
	if (remainder > 0)
		quotient++;

	return quotient;
}

static int get_brightness(int *val)
{
	int brt, ret;

	if (!display_dev || !display_dev->get_brightness) {
		_E("There is no display device.");
		return -ENOENT;
	}

	ret = display_dev->get_brightness(&brt);
	if (ret < 0) {
		_E("Failed to get brightness: %d", ret);
		return ret;
	}

	*val = get_brt_normalized(brt);
	return 0;
}

static int get_brightness_by_light_sensor(float lmax, float lmin, float light, int *brt)
{
	int brt_raw;
	int ret;

	if (!display_dev || !display_dev->get_auto_brightness)
		return -ENOTSUP;

	ret = display_dev->get_auto_brightness(lmax, lmin, light, &brt_raw);
	if (ret < 0) {
		_E("Failed to get brightness by light sensor: %d", ret);
		return ret;
	}

	*brt = get_brt_normalized(brt_raw);
	return 0;
}

static int get_image_effect(enum display_image_effect *effect)
{
	int ret;
	enum display_image_effect val;

	if (!display_dev || !display_dev->get_image_effect) {
		_E("There is no display device.");
		return -ENOENT;
	}

	ret = display_dev->get_image_effect(&val);
	if (ret < 0) {
		_E("Failed to get image effect: %d", ret);
		return ret;
	}

	*effect = val;

	return 0;
}

static int set_image_effect(enum display_image_effect effect)
{
	int ret;

	if (!display_dev || !display_dev->set_image_effect) {
		_E("There is no display device.");
		return -ENOENT;
	}

	ret = display_dev->set_image_effect(effect);
	if (ret < 0) {
		_E("Failed to set image effect: %d", ret);
		return ret;
	}

	return 0;
}

static int get_panel_mode(enum display_panel_mode *mode)
{
	int ret;
	enum display_panel_mode val;

	if (!display_dev || !display_dev->get_panel_mode) {
		_E("there is no display device");
		return -ENOENT;
	}

	ret = display_dev->get_panel_mode(&val);
	if (ret < 0) {
		_E("failed to get panel mode(%d)", ret);
		return ret;
	}

	*mode = val;

	return 0;
}

static int set_panel_mode(enum display_panel_mode mode)
{
	int ret;

	if (!display_dev || !display_dev->set_panel_mode) {
		_E("there is no display device");
		return -ENOENT;
	}

	ret = display_dev->set_panel_mode(mode);
	if (ret < 0) {
		_E("failed to set panel mode(%d)", ret);
		return ret;
	}

	return 0;
}

static int get_frame_rate(int *rate)
{
	if (!rate)
		return -EINVAL;

	if (!display_dev || !display_dev->get_frame_rate)
		return -ENOTSUP;

	return display_dev->get_frame_rate(rate);
}

static int set_frame_rate(int rate)
{
	int ret;
	static int fmin = -1, fmax = -1;

	if (!display_dev ||
		!display_dev->set_frame_rate)
		return -ENOTSUP;

	if (display_dev->get_min_frame_rate) {
		if (fmin < 0) {
			ret = display_dev->get_min_frame_rate(&fmin);
			if (ret < 0) {
				_E("Failed to get min frate rate: %d", ret);
				return ret;
			}
		}
		if (rate < fmin) {
			_E("Invalid rate(%d). (Valid rate: %d <= rate)", rate, fmin);
			return -EINVAL;
		}
	}

	if (display_dev->get_max_frame_rate) {
		if (fmax < 0) {
			ret = display_dev->get_max_frame_rate(&fmax);
			if (ret < 0) {
				_E("Failed to get max frate rate: %d", ret);
				return ret;
			}
		}
		if (rate > fmax) {
			_E("Invalid rate(%d). (Valid rate: rate <= %d)", rate, fmax);
			return -EINVAL;
		}
	}

	return display_dev->set_frame_rate(rate);
}

/* It was operated only AOD enter & leave */
static int backlight_transit_state(int state)
{
	int brt, val;
	int start, end;
	static int aod_brightness_level;

	backlight_ops.get_brightness(&brt);

	if (state == DPMS_OFF) {
		start = brt;

		/* There are three levels of AOD brightness: max, normal, min brightness.
		 * At first, AOD brightness is set to normal level when charger is connected,
		 * otherwise set to max. And then, if the current brightness is lower than this
		 * initialized AOD brightness, set AOD brightness to min level brightness. */
		if (battery.online_type && aod_normal_level > 0)
			aod_brightness_level = aod_normal_level;
		else if (aod_max_level > 0)
			aod_brightness_level = aod_max_level;
		else
			aod_brightness_level = display_conf.aod_enter_level;

		/*
		 * The value of backlight_ops.get_brightness is system brightness.
		 * But when device is LBM, the value is not same with real brightness.
		 * So it should be read exactly value for transit smooth effect
		 */
		get_brightness(&val);

		if ((val < aod_brightness_level) && (aod_min_level > 0))
			aod_brightness_level = aod_min_level;

		if (val > aod_brightness_level) {
			_D("backlight transit start %d end %d DPMS OFF", start, aod_brightness_level);
			backlight_ops.transit_brt(start, aod_brightness_level, display_conf.brightness_change_step);
		}
	} else {
		/* prevent transit effect when another effect is already executed */
		if (brt != aod_brightness_level) {
			_W("effect is already executed brt(%d) aod_level(%d)",
					brt, aod_brightness_level);
			return 0;
		}

		start = aod_brightness_level;
		end = default_brightness;
		_D("backlight transit start %d end %d DPMS ON", start, end);
		backlight_ops.transit_brt(start, end, display_conf.brightness_change_step);
	}

	return 0;
}

static gboolean blink_cb(gpointer data)
{
	static bool flag;

	set_brightness(flag ? PM_MAX_BRIGHTNESS : PM_MIN_BRIGHTNESS);

	flag = !flag;

	return G_SOURCE_CONTINUE;
}

static void blink(int timeout)
{
	static guint timer;

	if (timer) {
		g_source_remove(timer);
		timer = 0;
	}

	if (timeout < 0) {
		_E("timeout value is invalid %d", timeout);
		return;
	}

	if (timeout == 0) {
		backlight_update();
		return;
	}

	timer = g_timeout_add(timeout, blink_cb, NULL);
}

static gboolean release_blink_cb(gpointer data)
{
	blink(0);

	release_timer = 0;
	return G_SOURCE_REMOVE;
}

static void release_blink(void)
{
	if (release_timer) {
		g_source_remove(release_timer);
		release_timer = 0;
	}

	release_timer = g_timeout_add(DUMP_MODE_WAITING_TIME, release_blink_cb, NULL);
}

static void restore_brightness_func(void)
{
	backlight_ops.set_brightness = set_brightness;
	backlight_ops.get_brightness = get_brightness;
	backlight_ops.transit_brt = change_brightness;
}

static void _init_ops(void)
{
	backlight_ops.off = backlight_off;
	backlight_ops.dim = backlight_dim;
	backlight_ops.on = backlight_on;
	backlight_ops.update = backlight_update;
	backlight_ops.standby = backlight_standby;
	backlight_ops.set_default_brt = set_default_brt;
	backlight_ops.get_default_brt = get_default_brt;
	backlight_ops.get_lcd_power = get_lcd_power;
	backlight_ops.set_custom_status = set_custom_status;
	backlight_ops.get_custom_status = get_custom_status;
	backlight_ops.save_custom_brightness = save_custom_brightness;
	backlight_ops.custom_update = custom_backlight_update;
	backlight_ops.set_force_brightness = set_force_brightness;
	backlight_ops.set_brightness = set_brightness;
	backlight_ops.get_brightness = get_brightness;
	backlight_ops.restore_brightness_func = restore_brightness_func;
	backlight_ops.get_brightness_by_light_sensor = get_brightness_by_light_sensor;
	backlight_ops.get_image_effect = get_image_effect;
	backlight_ops.set_image_effect = set_image_effect;
	backlight_ops.get_panel_mode = get_panel_mode;
	backlight_ops.set_panel_mode = set_panel_mode;
	backlight_ops.get_frame_rate = get_frame_rate;
	backlight_ops.set_frame_rate = set_frame_rate;
	backlight_ops.transit_state = backlight_transit_state;
	backlight_ops.transit_brt = change_brightness;
	backlight_ops.blink = blink;
	backlight_ops.release_blink = release_blink;

	power_ops.suspend = system_suspend;
	power_ops.enable_autosleep = system_enable_autosleep;
	power_ops.power_lock = system_power_lock;
	power_ops.power_unlock = system_power_unlock;
	power_ops.get_power_lock = system_get_power_lock;
	power_ops.get_power_lock_support = system_get_power_lock_support;
	power_ops.check_wakeup_src = check_wakeup_src;
	power_ops.get_wakeup_count = get_wakeup_count;
	power_ops.set_wakeup_count = set_wakeup_count;
}

int display_service_load(void)
{
	struct hw_info *info;
	int r;

	if (display_dev)
		return 0;

	r = hw_get_info(DISPLAY_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (r < 0) {
		_I("display shared library is not supported: %d", r);
		return 0;
	}

	if (!info->open) {
		_E("Failed to open display device: open(NULL).");
		return -EPERM;
	}

	r = info->open(info, NULL, (struct hw_common **)&display_dev);
	if (r < 0) {
		_E("Failed to get display device structure: %d", r);
		return -EPERM;
	}

	if (display_dev && display_dev->get_aod_brightness) {
		r = display_dev->get_aod_brightness(&aod_max_level, &aod_normal_level, &aod_min_level);
		if (r == 0) {
			_I("Use various AOD brightness : max %d, normal %d, min %d", aod_max_level, aod_normal_level, aod_min_level);
		}
	}

	_D("Display device structure load success.");

	return 0;
}

int display_service_free(void)
{
	struct hw_info *info;

	if (!display_dev)
		return -ENOENT;

	info = display_dev->common.info;

	assert(info);

	info->close((struct hw_common *)display_dev);
	display_dev = NULL;

	return 0;
}

bool vital_mode(void)
{
	return vital_sleep;
}

bool is_display_detached(void)
{
	int state = backlight_ops.get_lcd_power();

	return (state == DPMS_DETACH) || (state == -EINVAL);
}

static int vital_state_changed(void *data)
{
	int type;

	assert(data);

	type = *(int *)data;
	if (type == VITAL_EXIT)
		suspend_other_process(VITAL_EXIT);

	return 0;
}

int init_sysfs(unsigned int flags)
{
	_init_ops();

	register_notifier(DEVICE_NOTIFIER_VITAL_STATE, vital_state_changed);

	return 0;
}

int exit_sysfs(void)
{
	int fd;
	const struct device_ops *ops = NULL;

	fd = open("/tmp/sem.pixmap_1", O_RDONLY);
	if (fd == -1) {
		_E("X server disable");
		backlight_on(NORMAL_MODE);
	}

	backlight_update();

	exit_dpms();

	ops = find_device("touchscreen");
	if (!check_default(ops))
		ops->start(NORMAL_MODE);

	ops = find_device("touchkey");
	if (!check_default(ops))
		ops->start(NORMAL_MODE);

	if (fd != -1)
		close(fd);

	unregister_notifier(DEVICE_NOTIFIER_VITAL_STATE, vital_state_changed);

	return 0;
}
