/*
 * test
 *
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <sys/time.h>

#include "test.h"
#include "test_dbus_interface.h"

#define METHOD_BATTERY_MONITOR_BASIC           "basic"
#define METHOD_BATTERY_MONITOR_ONOFF           "onoff"
#define METHOD_BATTERY_MONITOR_BRIGHTNESS      "brightness"
#define METHOD_BATTERY_MONITOR_BASIC_LCD_OFF   "basic_lcdoff"
#define METHOD_BATTERY_MONITOR_BASIC_LCD_ON    "basic_lcdon"
#define METHOD_BATTERY_MONITOR_BASIC_SESSIONS  "basic_sessions"


static struct timeval t_start, t_end;
static void time_set_start(void)
{
	gettimeofday(&t_start, NULL);
}

static void time_set_end(void)
{
	gettimeofday(&t_end, NULL);
}

static int time_get_duration_sec(void)
{
	return t_end.tv_sec - t_start.tv_sec;
}

#define DBUS_DEVICED "org.tizen.system.deviced"
#define DBUS_DEVICED_BM_PATH "/Org/Tizen/System/DeviceD/BatteryMonitor"
#define DBUS_DEVICED_BM_IFACE "org.tizen.system.deviced.BatteryMonitor"
#define DBUS_DEVICED_BM_MEMBER "GetBMData"

/*
 * common structure for application time map.
 */
typedef struct {
	char *app_id;		/**< application id */
	uint time;		/**< total duration for which application as active */
} app_time_map_st1;

/*
 * structure for "display" feature data.
 */
typedef struct {
	uint high;			/**< total time in milliseconds during which brightness intensity was high */
	uint low;			/**< total time in milliseconds during which brightness intensity was low */
	uint med;			/**< total time in milliseconds during which brightness intensity was medium */
	time_t start;			/**< start time of feature data collection session */
	time_t stop;			/**< stop time of feature data collection session */
	GSList *atm_list;		/**< application time map('app_time_map_st1') list for all active applications between start & stop */
	GSList *display_list;		/**< list of nodes of 'bm_display_st' for next consecutive sessions */
} bm_display_st;

static void _atm_list_free(gpointer data)
{
	app_time_map_st1 *atm = (app_time_map_st1 *)data;

	if (!atm) {
		_E("atm is null");
		return;
	}

	if (atm->app_id)
		g_free(atm->app_id);

	free(atm);
}

static void _bds_list_free(gpointer data)
{
	bm_display_st *bds = (bm_display_st *)data;

	if (!bds) {
		_E("bds is null");
		return;
	}

	if (bds->atm_list)
		g_slist_free_full(bds->atm_list, _atm_list_free);

	free(bds);
}

static void bds_header_free(bm_display_st *bds_header)
{
	if (bds_header) {
		g_slist_free_full(bds_header->display_list, _bds_list_free);
		free(bds_header);
	}
}

static bm_display_st *_variant_to_bds(GVariant *param)
{
	bm_display_st *bds_header = NULL;
	bm_display_st *bds = NULL;
	GSList *bds_list = NULL;
	app_time_map_st1 *atm = NULL;
	GVariantIter *iter = NULL;
	GVariantIter *sub_iter = NULL;
	GSList *atm_list = NULL;
	guint vals[3];
	guint64 t_start, t_end;
	gchar *app_id = NULL;
	guint app_time = 0;

	if (!param) {
		_E("Wrong param:null");
		return NULL;
	}

	g_variant_get(param, "(a(uuuxxa(su)))", &iter);
	if (g_variant_iter_n_children(iter) == 0) {
		_E("failed to get data from gvariant. array size is 0.");
		goto err;
	}

	/* make bds list */
	while (g_variant_iter_loop(iter, "(uuuxxa(su))",
					&vals[0], &vals[1], &vals[2], &t_start, &t_end,
					&sub_iter)) {
		/* make atm list */
		atm_list = NULL;
		while (g_variant_iter_loop(sub_iter, "(su)", &app_id, &app_time)) {
			atm = (app_time_map_st1 *)malloc(sizeof(app_time_map_st1));
			if (!atm) {
				_E("failed to allocate memory");
				goto err;
			}

			atm->app_id = g_strdup(app_id);
			atm->time = app_time;

			atm_list = g_slist_prepend(atm_list, atm);
		}

		bds = (bm_display_st *)malloc(sizeof(bm_display_st));
		if (!bds) {
			_E("failed to allocate memory");
			g_slist_free_full(atm_list, _atm_list_free);
			goto err;
		}

		bds->high = vals[0];
		bds->low = vals[1];
		bds->med = vals[2];
		bds->start = t_start;
		bds->stop = t_end;
		bds->atm_list = g_slist_reverse(atm_list);
		bds->display_list = NULL;

		bds_list = g_slist_prepend(bds_list, bds);
	}

	bds_header = (bm_display_st *)calloc(1, sizeof(bm_display_st));
	if (!bds_header) {
		_E("failed to allocate memory");
		goto err;
	}

	bds_header->display_list = g_slist_reverse(bds_list);

	g_variant_iter_free(iter);

	return bds_header;

err:
	if (iter)
		g_variant_iter_free(iter);
	if (bds_list)
		g_slist_free_full(bds_list, _bds_list_free);

	return NULL;
}

static bm_display_st *_get_display_info(void)
{
	GVariant *reply = NULL;
	bm_display_st *bds = NULL;

	reply = test_deviced_BatteryMonitor_GetBMData();
	if (!reply) {
		_E("failed to get battery monitor data");
		return NULL;
	}

	/* convert gvariant into bds list */
	bds = _variant_to_bds(reply);

	g_variant_unref(reply);

	return bds;
}

/* lcd off & reset monitor measurement data */
static bool _reset_battery_monitor_and_lcdoff(void)
{
	int ret;
	GVariant *reply;

	ret = test_deviced_display_CustomLCDOff("proximity");
	if (ret < 0) {
		_E("failed to get battery monitor data");
		return false;
	}

	/* reset monitor */
	reply = test_deviced_BatteryMonitor_GetBMData();
	if (!reply) {
		_E("failed to get battery monitor data");
		return false;
	}
	g_variant_unref(reply);

	return true;
}

static int _bds_get_total_brightness_time(bm_display_st *bds)
{
	if (!bds)
		return 0;

	return bds->high + bds->med + bds->low;
}

static int _bds_get_total_app_time(bm_display_st *bds)
{
	GSList *node;
	int sum = 0;

	if (!bds || !bds->atm_list)
		return 0;

	for (node = bds->atm_list; node != NULL; node = g_slist_next(node))
		sum += ((app_time_map_st1*)(node->data))->time;

	return sum;
}

static bool _bds_check_brightness_and_app_time(bm_display_st *bds_header)
{
	GSList *bds_list;
	bm_display_st *bds;
	int brightness;
	int app_time;

	if (!bds_header)
		return false;

	for (bds_list = bds_header->display_list; bds_list != NULL; bds_list = g_slist_next(bds_list)) {
		bds = (bm_display_st*)bds_list->data;
		brightness = _bds_get_total_brightness_time(bds);
		app_time = _bds_get_total_app_time(bds);

		if (brightness != app_time) {
			_E("bds total brightness time(%d) != total app time(%d)",
				brightness, app_time);
			return false;
		}
	}

	return true;
}

static int _bds_list_get_total_brightness_time_ms(bm_display_st *bds_header)
{
	GSList *bds_list;
	bm_display_st *bds;
	int sum = 0;

	if (!bds_header)
		return 0;

	for (bds_list = bds_header->display_list; bds_list != NULL; bds_list = g_slist_next(bds_list)) {
		bds = (bm_display_st*)bds_list->data;
		sum += _bds_get_total_brightness_time(bds);
	}

	_E("sum = %d", sum);

	return sum;
}

static void print_atm_list(GSList *atm_list)
{
	GSList *list;
	app_time_map_st1 *atm;

	for (list = atm_list; list != NULL; list = g_slist_next(list)) {
		atm = (app_time_map_st1 *)list->data;
		_I("%s %u", atm->app_id, atm->time);
	}
}

static void print_bds_info(bm_display_st *bds)
{
	_I("start:%ld, end:%ld, high:%d, med:%d, low:%d, atm_list len:%d",
			bds->start, bds->stop, bds->high, bds->med, bds->low, g_slist_length(bds->atm_list));
	print_atm_list(bds->atm_list);
}

/* TC */

/*
   basic dbus interface check
   result : reply data
*/
static bool test_battery_monitor_is_ready(void)
{
	GVariant *ret = test_deviced_BatteryMonitor_GetBMData();
	if (!ret) {
		_E("failed to get battery monitor data");
		return false;
	}
	g_variant_unref(ret);

	return true;
}

/*
   basic lcd on/off test
   result : reply single session data
*/
static bool test_battery_monitor_lcd_on_off(void)
{
	bool pass = true;
	int ret;
	int duration;
	bm_display_st *bds_header = NULL;
	bm_display_st *bds = NULL;
	int sum;

	if (!_reset_battery_monitor_and_lcdoff()) {
		_E("failed to get battery monitor data");
		return false;
	}

	ret = test_deviced_display_CustomLCDOn(1);
	if (ret < 0) {
		_E("failed to get battery monitor data");
		return false;
	}

	/* start test */
	time_set_start();
	sleep(1);
	time_set_end();

	ret = test_deviced_display_CustomLCDOff("proximity");
	if (ret < 0) {
		_E("failed to get battery monitor data");
		return false;
	}
	/* end test */

	/* check result */
	duration = time_get_duration_sec();

	bds_header = _get_display_info();
	if (!bds_header) {
		_E("failed to get battery monitor data");
		pass = false;
		goto err;
	}

	if (!_bds_check_brightness_and_app_time(bds_header)) {
		_E("failed to _bds_check_brightness_and_app_time");
		pass = false;
		goto err;
	}

	bds = (bm_display_st *)bds_header;
	sum = _bds_list_get_total_brightness_time_ms(bds_header);
	_I("duration=%d, bm=%d, start-end(%d)", duration, sum, (int)(bds->stop - bds->start));

	/* approximation : compare sec */
	if (duration != sum / 1000) {
		_E("duration mismatched");
		pass = false;
	}

err:
	bds_header_free(bds_header);

	return pass;
}

/*
   basic test of brightness(high/med/low)
   result : reply single session data and high/med/low value.
*/
static bool test_battery_monitor_brightness(void)
{
	bool pass = true;
	int ret;
	int duration;
	bm_display_st *bds_header = NULL;
	bm_display_st *bds = NULL;
	int sum;

	if (!_reset_battery_monitor_and_lcdoff()) {
		_E("failed to get battery monitor data");
		return false;
	}

	/* init test env */
	ret = test_deviced_display_SetBrightness(0, 100);
	if (ret < 0) {
		_E("failed to get battery monitor data");
		return false;
	}

	/* start test */

	/* high 1sec */
	ret = test_deviced_display_CustomLCDOn(1);
	if (ret < 0) {
		_E("failed to get battery monitor data");
		return false;
	}

	time_set_start();
	sleep(1);

	/* med 1sec */
	ret = test_deviced_display_SetBrightness(0, 40);
	if (ret < 0) {
		_E("failed to get battery monitor data");
		return false;
	}
	sleep(1);

	/* low 1sec */
	ret = test_deviced_display_SetBrightness(0, 10);
	if (ret < 0) {
		_E("failed to get battery monitor data");
		return false;
	}
	sleep(1);
	time_set_end();

	ret = test_deviced_display_CustomLCDOff("proximity");
	if (ret < 0) {
		_E("failed to get battery monitor data");
		return false;
	}
	/* end test */

	/* check result */
	duration = time_get_duration_sec();

	bds_header = _get_display_info();
	if (!bds_header) {
		_E("failed to get battery monitor data");
		return false;
	}

	/* expect single session data */
	if (g_slist_length(bds_header->display_list) != 1) {
		_E("size(display_list) should be 1, but(%d)", g_slist_length(bds_header->display_list));
		pass = false;
		goto err;
	}

	if (!_bds_check_brightness_and_app_time(bds_header)) {
		pass = false;
		goto err;
	}

	bds = (bm_display_st *)bds_header->display_list->data;
	sum = _bds_list_get_total_brightness_time_ms(bds_header);
	_I("duration=%d, bm=%d", duration, sum);

	if (duration != sum / 1000) {
		_E("sleep duration(%d) != h+l+m(%d)", duration, sum / 1000);
		pass = false;
	}
	if (bds->high / 1000 != 1) {
		_E("bds->high(%d) != 1", bds->high / 1000);
		pass = false;
	}
	if (bds->med / 1000 != 1) {
		_E("bds->high(%d) != 1", bds->med / 1000);
		pass = false;
	}
	if (bds->low / 1000 != 1) {
		_E("bds->high(%d) != 1", bds->low / 1000);
		pass = false;
	}

err:
	bds_header_free(bds_header);

	return pass;
}

/*
   basic test of multiple GetBMData call in lcd off state
   result : reply single session data(filled with 0)
*/
static bool test_battery_monitor_request_within_lcd_off(void)
{
	bm_display_st *bds_header = NULL;
	bm_display_st *bds = NULL;
	int i;

	if (!_reset_battery_monitor_and_lcdoff()) {
		_E("failed to get battery monitor data");
		return false;
	}

	for (i = 0; i < 10; ++i) {
		bds_header = _get_display_info();
		if (!bds_header) {
			_E("failed to get battery monitor data");
			return false;
		}
		/* expect single session */
		if (g_slist_length(bds_header->display_list) != 1) {
			_E("failed. size(display list) should be 1");
			goto err;
		}

		/* expect zero datas */
		bds = (bm_display_st *)bds_header->display_list->data;
		if (bds->start || bds->stop || bds->high || bds->med || bds->low || g_slist_length(bds->atm_list)) {
			_E("bds data should be filled with 0");
			print_bds_info(bds);
			goto err;
		}

		g_slist_free_full(bds_header->display_list, _bds_list_free);
		free(bds_header);
	}

	return true;
err:
	bds_header_free(bds_header);
	return false;
}

/*
   basic test of multiple GetBMData call in lcd on state
   result : reply single session data(filled with measurement data)
*/
static bool test_battery_monitor_request_within_lcd_on(void)
{
	bm_display_st *bds_header = NULL;
	bm_display_st *bds = NULL;
	int i;
	int ret;

	if (!_reset_battery_monitor_and_lcdoff()) {
		_E("failed to get battery monitor data");
		return false;
	}

	/* lcd on */
	ret = test_deviced_display_CustomLCDOn(1);
	if (ret < 0) {
		_E("failed to get battery monitor data");
		return false;
	}

	/* test */
	for (i = 0; i < 10; ++i) {
		bds_header = _get_display_info();
		if (!bds_header) {
			_E("failed to get battery monitor data");
			return false;
		}
		/* expect single session */
		if (g_slist_length(bds_header->display_list) != 1) {
			_E("failed. size(display list) should be 1");
			goto err;
		}

		/* expect some measurement datas */
		bds = (bm_display_st *)bds_header->display_list->data;
		if (!bds->start || !bds->stop || !(bds->high + bds->med + bds->low) || !g_slist_length(bds->atm_list)) {
			_E("bds data should be filled with datas");
			print_bds_info(bds);
			goto err;
		}

		g_slist_free_full(bds_header->display_list, _bds_list_free);
		free(bds_header);
	}

	return true;
err:
	bds_header_free(bds_header);

	return false;
}

/*
   battery monitor limitation test ?
   result : reply 100 session datas(filled with measurement data)
*/
static bool test_battery_monitor_generate_many_sessions(void)
{
	bm_display_st *bds_header = NULL;
	bm_display_st *bds = NULL;
	GSList *list;
	const int session_size = 100;
	int i;
	int ret;

	if (!_reset_battery_monitor_and_lcdoff()) {
		_E("failed to get battery monitor data");
		return false;
	}

	for (i = 0; i < session_size; ++i) {
		/* lcd on */
		ret = test_deviced_display_CustomLCDOn(1);
		if (ret < 0) {
			_E("failed to get battery monitor data");
			return false;
		}
		usleep(1000);
		ret = test_deviced_display_CustomLCDOff("proximity");
		if (ret < 0) {
			_E("failed to get battery monitor data");
			return false;
		}
	}

	/* test */
	bds_header = _get_display_info();
	if (!bds_header) {
		_E("failed to get battery monitor data");
		return false;
	}

	if (g_slist_length(bds_header->display_list) != session_size) {
		_E("failed. size(display list) should be %d. but %d", session_size, g_slist_length(bds_header->display_list));
		goto err;
	}

	/* expect measurement datas */
	for (list = bds_header->display_list; list != NULL; list = g_slist_next(list)) {
		bds = (bm_display_st *)list->data;
		if (!bds->start || !bds->stop || !(bds->high + bds->med + bds->low) || !g_slist_length(bds->atm_list)) {
			_E("bds data should be filled with datas");
			print_bds_info(bds);
			goto err;
		}
	}

err:
	bds_header_free(bds_header);

	return true;
}

void battery_monitor_test_all(int *success, int *fail)
{
	int s = 0;
	int f = 0;

	(test_battery_monitor_is_ready())    ? s++ : f++;
	(test_battery_monitor_lcd_on_off())    ? s++ : f++;
	(test_battery_monitor_brightness())    ? s++ : f++;
	//(test_battery_monitor_time_apps())    ? s++ : f++;
	(test_battery_monitor_request_within_lcd_off())    ? s++ : f++;
	(test_battery_monitor_request_within_lcd_on())    ? s++ : f++;
	(test_battery_monitor_generate_many_sessions())    ? s++ : f++;

	if (success)
		*success = s;
	if (fail)
		*fail = f;
}

static void battery_monitor_init(void *data)
{
	int success = 0;
	int fail = 0;

	_I("Start test.");

	battery_monitor_test_all(&success, &fail);

	_I("Total=%d Success=%d Fail=%d", success+fail, success, fail);
}

static void battery_monitor_exit(void *data)
{
	_I("End test.");
}

/* systemctl stop batterymonitor.service */
static int battery_monitor_unit(int argc, char **argv)
{
	if (argc < 4) {
		int success = 0;
		int fail = 0;
		_I("Start test.");
		battery_monitor_test_all(&success, &fail);
		_I("Total=%d Success=%d Fail=%d", success+fail, success, fail);
	} else if (0 == strcasecmp(argv[3], METHOD_BATTERY_MONITOR_BASIC)) {
		test_battery_monitor_is_ready();
	} else if (0 == strcasecmp(argv[3], METHOD_BATTERY_MONITOR_ONOFF)) {
		test_battery_monitor_lcd_on_off();
	} else if (0 == strcasecmp(argv[3], METHOD_BATTERY_MONITOR_BRIGHTNESS)) {
		test_battery_monitor_brightness();
	} else if (0 == strcasecmp(argv[3], METHOD_BATTERY_MONITOR_BASIC_LCD_OFF)) {
		test_battery_monitor_request_within_lcd_off();
	} else if (0 == strcasecmp(argv[3], METHOD_BATTERY_MONITOR_BASIC_LCD_ON)) {
		test_battery_monitor_request_within_lcd_on();
	} else if (0 == strcasecmp(argv[3], METHOD_BATTERY_MONITOR_BASIC_SESSIONS)) {
		test_battery_monitor_generate_many_sessions();
	} else {
		_E("Unknown test case.");
	}

	return 0;
}

static const struct test_ops battery_monitor_test_ops = {
	.priority = TEST_PRIORITY_NORMAL,
	.name     = "battery-monitor",
	.init     = battery_monitor_init,
	.exit     = battery_monitor_exit,
	.unit     = battery_monitor_unit,
};

TEST_OPS_REGISTER(&battery_monitor_test_ops)
