/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include "log.h"
#include "device-notifier.h"
#include "list.h"
#include "common.h"
#include "core/devices.h"
#include <libsyscommon/dbus-system.h>

struct device_notifier {
	bool deleted;
	enum device_notifier_type type;
	int (*func)(void *data);
};

static dd_list *device_notifier_list;
static guint idl;
static guint late_init_timer;

#define LATE_INIT_WAIT_TIME	30   /* second */
#define DEFAULT_LATE_INIT_VALUE	(guint)0x0DEF0DEF

#define FIND_NOTIFIER(a, b, d, e, f) \
	DD_LIST_FOREACH(a, b, d) \
		if (e == d->e && f == (d->f))

#define NOTIFY_STR(x)	[(x)] = __stringify((x))

static const char *device_notifier_type_str[DEVICE_NOTIFIER_MAX] = {
	NOTIFY_STR(DEVICE_NOTIFIER_DAEMON_RESTARTED),
	NOTIFY_STR(DEVICE_NOTIFIER_BOOTING_DONE),
	NOTIFY_STR(DEVICE_NOTIFIER_LCD),
	NOTIFY_STR(DEVICE_NOTIFIER_LCD_OFF),
	NOTIFY_STR(DEVICE_NOTIFIER_LOWBAT),
	NOTIFY_STR(DEVICE_NOTIFIER_FULLBAT),
	NOTIFY_STR(DEVICE_NOTIFIER_POWER_SUPPLY),
	NOTIFY_STR(DEVICE_NOTIFIER_BATTERY_HEALTH),
	NOTIFY_STR(DEVICE_NOTIFIER_BATTERY_PRESENT),
	NOTIFY_STR(DEVICE_NOTIFIER_BATTERY_OVP),
	NOTIFY_STR(DEVICE_NOTIFIER_BATTERY_CHARGING),
	NOTIFY_STR(DEVICE_NOTIFIER_DISPLAY_AMBIENT_CONDITION),
	NOTIFY_STR(DEVICE_NOTIFIER_DISPLAY_AMBIENT_STATE),
	NOTIFY_STR(DEVICE_NOTIFIER_DISPLAY_LOCK),
	NOTIFY_STR(DEVICE_NOTIFIER_POWER_RESUME),
	NOTIFY_STR(DEVICE_NOTIFIER_POWEROFF),
	NOTIFY_STR(DEVICE_NOTIFIER_PROCESS_BACKGROUND),
	NOTIFY_STR(DEVICE_NOTIFIER_PROCESS_FOREGROUND),
	NOTIFY_STR(DEVICE_NOTIFIER_USB_DEBUG_MODE),
	NOTIFY_STR(DEVICE_NOTIFIER_USB_TETHERING_MODE),
	NOTIFY_STR(DEVICE_NOTIFIER_EVENT_HANDLER),
	NOTIFY_STR(DEVICE_NOTIFIER_EARLY_BOOTING_DONE),
	NOTIFY_STR(DEVICE_NOTIFIER_PMQOS),
	NOTIFY_STR(DEVICE_NOTIFIER_PMQOS_ULTRAPOWERSAVING),
	NOTIFY_STR(DEVICE_NOTIFIER_PMQOS_POWERSAVING),
	NOTIFY_STR(DEVICE_NOTIFIER_COOL_DOWN),
	NOTIFY_STR(DEVICE_NOTIFIER_VITAL_STATE),
	NOTIFY_STR(DEVICE_NOTIFIER_LONGKEY_RESTORE),
	NOTIFY_STR(DEVICE_NOTIFIER_UPSM),
	NOTIFY_STR(DEVICE_NOTIFIER_UPSM_OFF),
	NOTIFY_STR(DEVICE_NOTIFIER_BEZEL_WAKEUP),
	NOTIFY_STR(DEVICE_NOTIFIER_DISPLAY_BRIGHTNESS),
	NOTIFY_STR(DEVICE_NOTIFIER_CRITICAL_LOG),
	NOTIFY_STR(DEVICE_NOTIFIER_ULTRAPOWERSAVING),
};

int __register_notifier(enum device_notifier_type type, notify_cb func, const char *caller)
{
	dd_list *n;
	struct device_notifier *notifier;

	_I("%s, %p by %s", device_notifier_type_str[type], func, caller);

	if (!func) {
		_E("invalid func address! by %s", caller);
		return -EINVAL;
	}

	FIND_NOTIFIER(device_notifier_list, n, notifier, type, func) {
		_E("function is already registered! [%s, %p] by %s",
		    device_notifier_type_str[type], func, caller);
		return -EINVAL;
	}

	notifier = calloc(1, sizeof(struct device_notifier));
	if (!notifier) {
		_E("Fail to malloc for %s notifier! by %s", device_notifier_type_str[type], caller);
		return -ENOMEM;
	}

	notifier->type = type;
	notifier->func = func;

	DD_LIST_APPEND(device_notifier_list, notifier);

	return 0;
}

int __unregister_notifier(enum device_notifier_type type, notify_cb func, const char *caller)
{
	dd_list *n;
	struct device_notifier *notifier;

	if (!func) {
		_E("invalid func address of %s! by %s", device_notifier_type_str[type], caller);
		return -EINVAL;
	}

	FIND_NOTIFIER(device_notifier_list, n, notifier, type, func) {
		_I("[%s, %p] by %s", device_notifier_type_str[type], func, caller);
		notifier->deleted = true;
	}

	return 0;
}

static gboolean delete_unused_notifier_cb(void *data)
{
	dd_list *n;
	dd_list *next;
	struct device_notifier *notifier;

	DD_LIST_FOREACH_SAFE(device_notifier_list, n, next, notifier) {
		if (notifier->deleted) {
			DD_LIST_REMOVE_LIST(device_notifier_list, n);
			free(notifier);
		}
	}

	idl = 0;
	return G_SOURCE_REMOVE;
}

void device_notify(enum device_notifier_type type, void *data)
{
	dd_list *n;
	struct device_notifier *notifier;

	DD_LIST_FOREACH(device_notifier_list, n, notifier) {
		if (!notifier->deleted && type == notifier->type) {
			if (notifier->func)
				notifier->func(data);
		}
	}

	if (!idl)
		idl = g_idle_add(delete_unused_notifier_cb, NULL);
}

void device_notify_once(enum device_notifier_type type, void *data)
{
	dd_list *n;
	struct device_notifier *notifier;

	DD_LIST_FOREACH(device_notifier_list, n, notifier) {
		if (!notifier->deleted && type == notifier->type) {
			if (notifier->func)
				notifier->func(data);

			notifier->deleted = true;
		}
	}

	if (!idl)
		idl = g_idle_add(delete_unused_notifier_cb, NULL);
}

static void late_init_stop(void)
{
	if (late_init_timer == 0 || late_init_timer == DEFAULT_LATE_INIT_VALUE)
		return;

	g_source_remove(late_init_timer);
	late_init_timer = 0;
}

static int booting_done(void *data)
{
	static int done;

	if (data == NULL)
		goto out;

	done = *(int *)data;
	if (late_init_timer == 0)
		return done;
	late_init_stop();
out:
	return done;
}

static int early_booting_done(void *data)
{
	static int done;

	if (data == NULL)
		goto out;

	done = *(int *)data;
out:
	return done;
}

static gboolean late_init_timer_cb(void *data)
{
	int done;

	late_init_stop();
	done = early_booting_done(NULL);
	if (!done)
		device_notify(DEVICE_NOTIFIER_EARLY_BOOTING_DONE, (void *)&done);
	done = booting_done(NULL);
	late_init_timer = 0;
	if (done)
		return G_SOURCE_REMOVE;
	_I("late booting done");
	done = TRUE;
	device_notify(DEVICE_NOTIFIER_BOOTING_DONE, (void *)&done);
	return G_SOURCE_REMOVE;
}

static void device_notifier_init(void *data)
{
	int ret;

	ret = check_systemd_active();
	if (ret == TRUE) {
		_I("restart booting done");
		return;
	}
	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	register_notifier(DEVICE_NOTIFIER_EARLY_BOOTING_DONE, early_booting_done);
	late_init_timer = g_timeout_add_seconds(LATE_INIT_WAIT_TIME,
						late_init_timer_cb, NULL);

	if (!late_init_timer)
		late_init_timer = DEFAULT_LATE_INIT_VALUE;
}

static void device_notifier_exit(void *data)
{

}

static const struct device_ops notifier_device_ops = {
	DECLARE_NAME_LEN("notifier"),
	.init     = device_notifier_init,
	.exit     = device_notifier_exit,
};

DEVICE_OPS_REGISTER(&notifier_device_ops)

