/*
 * deviced
 *
 * Copyright (c) 2012 - 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdbool.h>
#include <fcntl.h>
#include <libsyscommon/dbus-system.h>

#include "util.h"
#include "core.h"
#include "display.h"
#include "display-ops.h"
#include "core/device-notifier.h"
#include "core/devices.h"

#define ON		"on"
#define OFF		"off"

#define SIGNAL_HOMESCREEN       "HomeScreen"

#define SIGNAL_ALPMLCDOFF		"ALPMLCDOff"
#define SIGNAL_ALPM_ON			"ALPMOn"
#define SIGNAL_ALPM_OFF			"ALPMOff"

#define CLOCK_START			"clockstart"
#define CLOCK_END			"clockend"
#define CLOCK_CHANGED			"clockchanged"
#define TIMEOUT_NONE			(-1)
#define AMBIENT_CLOCK_WAITING_TIME	5000 /* ms */

static int ambient_state;
static int ambient_condition;	/* Setting Value */
static pid_t ambient_pid;  /* Ambient Clock pid */
static unsigned int update_count;

void broadcast_ambient_state(int state)
{
	int ret;
	char *signal;

	signal = (state == true ? SIGNAL_ALPM_ON : SIGNAL_ALPM_OFF);
	ret = dbus_handle_emit_dbus_signal(NULL,
						DEVICED_PATH_DISPLAY,
						DEVICED_INTERFACE_DISPLAY,
						signal,
						NULL);
	if (ret < 0)
		_E("Failed to send dbus signal(%s).", signal);
}

int ambient_get_condition(void)
{
	return ambient_condition;
}

int ambient_get_state(void)
{
	return ambient_state;
}

static void ambient_set_condition(keynode_t *key_nodes, void *data)
{
	int state, val;

	if (key_nodes == NULL) {
		_E("Wrong parameter, key_nodes is null.");
		return;
	}

	val = vconf_keynode_get_bool(key_nodes);
	if (val != ambient_condition) {
		if (backlight_ops.get_lcd_power() != DPMS_ON)
			if (disp_plgn.pm_lock_internal)
				disp_plgn.pm_change_internal(INTERNAL_LOCK_PM, LCD_NORMAL);
	}

	ambient_condition = val;
	_I("Ambient mode condition is %d.", ambient_condition);

	state = (ambient_condition == 0 ? 0 : 1);
	device_notify(DEVICE_NOTIFIER_DISPLAY_AMBIENT_CONDITION, (void *)&state);

	set_dim_state(state);
}

int ambient_set_state(int on)
{
	if (!ambient_condition)
		return 0;

	broadcast_ambient_state(on);

	update_count = 0;

	/**
	 * When display is turned off before going to AOD, Generally pm receives
	 * clock signal or ALPMLCDOFF signal and decides to go to AOD or not.
	 * But a specific case, the signal is not received.
	 * So at that time deviced should turn off display to match the pair.
	 */
	if (on) {
		if (disp_plgn.pm_lock_internal)
			disp_plgn.pm_lock_internal(INTERNAL_LOCK_AMBIENT,
				LCD_OFF, STAY_CUR_STATE, AMBIENT_CLOCK_WAITING_TIME);
	} else
		ambient_pid = 0;

	_D("AMBIENT is %s.", (on ? ON : OFF));

	ambient_state = on;

	device_notify(DEVICE_NOTIFIER_DISPLAY_AMBIENT_STATE, (void *)&ambient_state);

	return 0;
}

void ambient_check_invalid_state(pid_t pid)
{
	if (pid != INTERNAL_LOCK_AMBIENT)
		return;

	if (is_emulator()) {
		/* In emulator, deviced does not turn off the display. */
		if (backlight_ops.get_lcd_power() == DPMS_ON)
			return;
	}

	if (ambient_get_state() == false)
		return;

	_I("Invalid state. Ambient state is change to off.");

	/* If lcd_power is on and ambient state is true
	 * when pm state is changed to sleep,
	 * deviced doesn't get the clock signal.
	 * deviced just turns off lcd in this case.
	 */

	reset_timeout(TIMEOUT_NONE);
	ambient_set_state(false);
	lcd_direct_control(DPMS_OFF, NORMAL_MODE);

	broadcast_lcd_off_late(LCD_OFF_LATE_MODE);
}

static void ambient_start_clock(void)
{
	if (pm_cur_state == S_NORMAL ||
	    pm_cur_state == S_LCDDIM ||
	    ambient_state == false)
		return;

	if (disp_plgn.pm_lock_internal)
		disp_plgn.pm_lock_internal(INTERNAL_LOCK_AMBIENT, LCD_OFF, STAY_CUR_STATE,
			AMBIENT_CLOCK_WAITING_TIME);
}

static void ambient_end_clock(pid_t pid)
{
	if (pm_cur_state == S_NORMAL ||
	    pm_cur_state == S_LCDDIM ||
	    ambient_state == false)
		return;

	if (update_count == 0) {
		_D("lcd off");

		backlight_ops.off(NORMAL_MODE);
		broadcast_lcd_off_late(LCD_OFF_LATE_MODE);
		if (disp_plgn.pm_unlock_internal)
			disp_plgn.pm_unlock_internal(INTERNAL_LOCK_AMBIENT, LCD_OFF, PM_SLEEP_MARGIN);
	}

	update_count++;
	if (update_count == UINT_MAX)
		update_count = 1;
	ambient_pid = pid;

	_I("Enter real ambient state by process(%d).",
	    ambient_pid);
}

static void homescreen_signal_handler(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)
{
	char *screen;
	pid_t pid;

	g_variant_get(param, "(s)", &screen);

	_D("screen : %s", screen);

	if (!screen)
		return;

	pid = dbus_connection_get_sender_pid(conn, sender);

	if (!strncmp(screen, CLOCK_START, strlen(CLOCK_START)))
		ambient_start_clock();
	else if (!strncmp(screen, CLOCK_END, strlen(CLOCK_END)))
		ambient_end_clock(pid);

	g_free(screen);
}

static void ambient_lcdoff_signal_handler(GDBusConnection  *conn,
	const gchar      *sender,
	const gchar      *path,
	const gchar      *iface,
	const gchar      *name,
	GVariant         *param,
	gpointer          data)
{
	if (ambient_state == false) {
		_E("It is not alpm mode.");
		return;
	}

	if (disp_plgn.pm_lock_internal)
		disp_plgn.pm_lock_internal(INTERNAL_LOCK_AMBIENT, LCD_OFF, GOTO_STATE_NOW, 0);

	_I("Display off in suspend state.");

	ambient_set_state(false);
	lcd_direct_control(DPMS_OFF, NORMAL_MODE);

	broadcast_lcd_off_late(LCD_OFF_LATE_MODE);
	if (disp_plgn.pm_unlock_internal)
		disp_plgn.pm_unlock_internal(INTERNAL_LOCK_AMBIENT, LCD_OFF, PM_SLEEP_MARGIN);
}

static void ambient_init(void *data)
{
	int ret;

	ret = vconf_get_bool("db/starter/always_on_display", &ambient_condition);
	if (ret < 0) {
		_E("Failed to get initial AOD condition. AOD is not going to work.");
		return;
	} else {
		_I("Ambient mode condition is %d.", ambient_condition);

		vconf_notify_key_changed("db/starter/always_on_display",
			ambient_set_condition, NULL);
	}
	ret = subscribe_dbus_signal(NULL,
				DEVICED_OBJECT_PATH,
				DEVICED_INTERFACE_NAME,
				SIGNAL_HOMESCREEN,
				homescreen_signal_handler,
				NULL, NULL);
	if (ret <= 0)
		_E("Failed to register signal handler: %d", ret);

	ret = subscribe_dbus_signal(NULL,
				DEVICED_OBJECT_PATH,
				DEVICED_INTERFACE_NAME,
				SIGNAL_ALPMLCDOFF,
				ambient_lcdoff_signal_handler,
				NULL, NULL);
	if (ret <= 0)
		_E("Failed to register signal handler: %d", ret);
}

static void ambient_exit(void *data)
{
	ambient_set_state(false);
	vconf_ignore_key_changed("db/starter/always_on_display", ambient_set_condition);
}

static const struct display_ops ambient_ops = {
	.name     = "ambient",
	.init     = ambient_init,
	.exit     = ambient_exit,
};

DISPLAY_OPS_REGISTER(&ambient_ops)
