/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <malloc.h>
#include <stdlib.h>
#include <unistd.h>
#include <poll.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/signal.h>
#include <fcntl.h>
#include <sys/eventfd.h>
#include <pthread.h>
#include <sys/ioctl.h>

#include <linux/usb/functionfs.h>
#include <systemd/sd-daemon.h>

#define FFS_PATH "/tmp/usb-host-test-ffs"

enum {
	EP0_IDX = 0,
	EP_IN_BULK_IDX,
	EP_OUT_BULK_IDX,
	EP_IN_INT_IDX,
	EP_OUT_INT_IDX,
	NUM_ENDPOINTS,
};

#define MAX(a, b) ((a) > (b) ? (a) : (b))

#define report_error(...) do {			\
		fprintf(stderr, __VA_ARGS__);	\
		putchar('\n');			\
	} while (0)

#define BUFSIZE 512
char int_buf[BUFSIZE];
char bulk_buf[BUFSIZE];

pthread_cond_t enable_cond = PTHREAD_COND_INITIALIZER;
pthread_cond_t disable_cond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t ready_lock;
int connected = 0;

/* Close all ep files */
void cleanup_ffs(int *ep)
{
	int i;

	for (i = 0; i < NUM_ENDPOINTS; ++i)
		close(ep[i]);
}

/* Handle events generated by kernel and provided via ep0 */
void *loop_ep0(void *data)
{
	struct usb_functionfs_event event;
	int ret;
	int *ep = data;

	while (1) {
		ret = read(ep[0], &event, sizeof(event));
		if (!ret) {
			report_error("Unable to read event from ep0.");
			return NULL;
		}

		switch (event.type) {
		case FUNCTIONFS_SETUP:
			/* stall for all setuo requests */
			if (event.u.setup.bRequestType & USB_DIR_IN) {
				if (write(ep[0], NULL, 0) < 0)
					report_error("write error: %d", errno);
			} else {
				if (read(ep[0], NULL, 0) < 0)
					report_error("read error: %d", errno);
			}

			fprintf(stderr, "Functionfs setup request\n");
			break;

		case FUNCTIONFS_ENABLE:
			usleep(200000);
			pthread_mutex_lock(&ready_lock);
			fprintf(stderr, "Functionfs enable request\n");
			if (connected) {
				report_error("Already enabled.\n");
				pthread_cond_broadcast(&disable_cond);
			}

			connected = 1;
			pthread_cond_broadcast(&enable_cond);
			pthread_mutex_unlock(&ready_lock);
			break;

		case FUNCTIONFS_DISABLE:
			pthread_mutex_lock(&ready_lock);
			fprintf(stderr, "Functionfs disable request\n");
			if (!connected) {
				report_error("Already disabled.\n");
				pthread_cond_broadcast(&enable_cond);
			}

			connected = 0;
			pthread_cond_broadcast(&disable_cond);
			pthread_mutex_unlock(&ready_lock);
			break;

		case FUNCTIONFS_BIND:
			fprintf(stderr, "Functionfs bind request\n");
			break;

		default:
			fprintf(stderr, "Functionf unknown request: %d\n", event.type);
			break;
		}
	}

	return NULL;
}

struct tdata {
	int ep_in;
	int ep_out;
	char *buf;
};

void *loop_thread(void *data)
{
	struct tdata *d = data;
	int ret;

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

	while (1) {
		ret = read(d->ep_out, d->buf, BUFSIZE);
		if (ret < 0) {
			report_error("Read error on fd %d: %m\n", d->ep_out);
			usleep(200000);
			if (errno != ECONNRESET && errno != ESHUTDOWN && errno != EINTR)
				exit(ret);

			pthread_exit(NULL);
		}

		fprintf(stderr, "received %d bytes\n", ret);
		usleep(200000);

		ret = write(d->ep_in, d->buf, ret);
		if (ret < 0) {
			report_error("Write error on fd %d: %m\n", d->ep_in);
			if (errno != ECONNRESET && errno != ESHUTDOWN && errno != EINTR)
				exit(ret);

			pthread_exit(NULL);
		}

		fprintf(stderr, "sent %d bytes\n", ret);
	}

	return NULL;
}

void clear_halt(int ep)
{
	int ret;

	ret = ioctl(ep, FUNCTIONFS_CLEAR_HALT);
	if (ret < 0)
		fprintf(stderr, "could not reset endpoint %d: %m", ep);
}

/* main chat function */
void do_chat(int *ep)
{
	int ret;
	pthread_t int_thread, bulk_thread, ep0_thread;
	struct tdata int_data, bulk_data;

	int_data.ep_in = ep[EP_IN_INT_IDX];
	int_data.ep_out = ep[EP_OUT_INT_IDX];
	int_data.buf = int_buf;
	bulk_data.ep_in = ep[EP_IN_BULK_IDX];
	bulk_data.ep_out = ep[EP_OUT_BULK_IDX];
	bulk_data.buf = bulk_buf;

	pthread_mutex_init(&ready_lock, NULL);

	ret = pthread_create(&ep0_thread, NULL, loop_ep0, ep);
	if (ret < 0)
		exit(1);

	while (1) {
		pthread_mutex_lock(&ready_lock);
		if (!connected)
			pthread_cond_wait(&enable_cond, &ready_lock);

		fprintf(stderr, "Connected\n");

		ret = pthread_create(&bulk_thread, NULL, loop_thread, &bulk_data);
		if (ret < 0)
			exit(1);

		ret = pthread_create(&int_thread, NULL, loop_thread, &int_data);
		if (ret < 0)
			exit(1);

		pthread_cond_wait(&disable_cond, &ready_lock);
		fprintf(stderr, "Disconnected\n");

		pthread_cancel(bulk_thread);
		pthread_cancel(int_thread);
		pthread_join(bulk_thread, NULL);
		pthread_join(int_thread, NULL);

		clear_halt(ep[EP_IN_BULK_IDX]);
		clear_halt(ep[EP_OUT_BULK_IDX]);
		clear_halt(ep[EP_IN_INT_IDX]);
		clear_halt(ep[EP_OUT_INT_IDX]);

		pthread_mutex_unlock(&ready_lock);
	}

	pthread_join(ep0_thread, NULL);
	pthread_mutex_destroy(&ready_lock);
}

int main(int argc, char **argv)
{
	int ep[NUM_ENDPOINTS];
	int i = 0;

	if (sd_listen_fds(0) < NUM_ENDPOINTS) {
		fprintf(stderr, "Expected endpoints not retrieved\n");
		return -1;
	}

	for (i = 0; i < NUM_ENDPOINTS; ++i)
		ep[i] = SD_LISTEN_FDS_START + i;

	do_chat(ep);
	cleanup_ffs(ep);

	return 0;
}
